"""
Simple integrators for the radiative transfer equation

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
cimport numpy as np
cimport cython
from stdlib cimport malloc, free, abs

cdef inline int imax(int i0, int i1):
    if i0 > i1: return i0
    return i1

cdef inline np.float64_t fmax(np.float64_t f0, np.float64_t f1):
    if f0 > f1: return f0
    return f1

cdef inline int imin(int i0, int i1):
    if i0 < i1: return i0
    return i1

cdef inline np.float64_t fmin(np.float64_t f0, np.float64_t f1):
    if f0 < f1: return f0
    return f1

cdef inline int iclip(int i, int a, int b):
    if i < a: return a
    if i > b: return b
    return i

cdef inline np.float64_t fclip(np.float64_t f,
                      np.float64_t a, np.float64_t b):
    return fmin(fmax(f, a), b)

cdef extern from "math.h":
    double exp(double x)
    float expf(float x)
    double floor(double x)
    double ceil(double x)
    double fmod(double x, double y)
    double log2(double x)
    long int lrint(double x)

cdef extern from "FixedInterpolator.h":
    np.float64_t fast_interpolate(int ds[3], int ci[3], np.float64_t dp[3],
                                  np.float64_t *data)
    np.float64_t offset_interpolate(int ds[3], np.float64_t dp[3], np.float64_t *data)
    np.float64_t trilinear_interpolate(int ds[3], int ci[3], np.float64_t dp[3],
                                       np.float64_t *data)
    np.float64_t eval_gradient(int *ds, int *ci, np.float64_t *dp,
                                       np.float64_t *data, np.float64_t *grad)

cdef class VectorPlane

cdef struct FieldInterpolationTable:
    # Note that we make an assumption about retaining a reference to values
    # externally.
    np.float64_t *values 
    np.float64_t bounds[2]
    np.float64_t dbin
    np.float64_t idbin
    int field_id
    int weight_field_id
    int weight_table_id
    int nbins
    int pass_through

cdef void FIT_initialize_table(FieldInterpolationTable *fit, int nbins,
              np.float64_t *values, np.float64_t bounds1, np.float64_t bounds2,
              int field_id, int weight_field_id = -1, int weight_table_id = -1,
              int pass_through = 0):
    fit.bounds[0] = bounds1; fit.bounds[1] = bounds2
    fit.nbins = nbins
    fit.dbin = (fit.bounds[1] - fit.bounds[0])/fit.nbins
    fit.idbin = 1.0/fit.dbin
    # Better not pull this out from under us, yo
    fit.values = values
    fit.field_id = field_id
    fit.weight_field_id = weight_field_id
    fit.weight_table_id = weight_table_id
    fit.pass_through = pass_through

cdef np.float64_t FIT_get_value(FieldInterpolationTable *fit,
                            np.float64_t *dvs):
    cdef np.float64_t bv, dy, dd, tf
    cdef int bin_id
    if fit.pass_through == 1: return dvs[fit.field_id]
    bin_id = <int> ((dvs[fit.field_id] - fit.bounds[0]) * fit.idbin)
    dd = dvs[fit.field_id] - (fit.bounds[0] + bin_id * fit.dbin) # x - x0
    if bin_id > fit.nbins - 2 or bin_id < 0: return 0.0
    bv = fit.values[bin_id]
    dy = fit.values[bin_id + 1] - bv
    if fit.weight_field_id != -1:
        return dvs[fit.weight_field_id] * (bv + dd*dy*fit.idbin)
    return (bv + dd*dy*fit.idbin)

cdef class TransferFunctionProxy:
    cdef int n_fields
    cdef int n_field_tables
    cdef public int ns

    # These are the field tables and their affiliated storage.
    # We have one field_id for every table.  Note that a single field can
    # correspond to multiple tables, and each field table will only have
    # interpolate called once.
    cdef FieldInterpolationTable field_tables[6]
    cdef np.float64_t istorage[6]

    # Here are the field tables that correspond to each of the six channels.
    # We have three emission channels, three absorption channels.
    cdef int field_table_ids[6]

    # We store a reference to the transfer function object and to the field
    # interpolation tables
    cdef public object tf_obj
    cdef public object my_field_tables

    def __cinit__(self, tf_obj):
        # We have N fields.  We have 6 channels.  We have M field tables.
        # The idea is that we can have multiple channels corresponding to the
        # same field table.  So, we create storage for the outputs from all the
        # field tables.  We need to know which field value to pass in to the
        # field table, and we need to know which table to use for each of the
        # six channels.
        cdef int i
        cdef np.ndarray[np.float64_t, ndim=1] temp
        cdef FieldInterpolationTable fit

        self.tf_obj = tf_obj

        self.n_field_tables = tf_obj.n_field_tables
        for i in range(6): self.istorage[i] = 0.0

        self.my_field_tables = []
        for i in range(self.n_field_tables):
            temp = tf_obj.tables[i].y
            FIT_initialize_table(&self.field_tables[i],
                      temp.shape[0],
                      <np.float64_t *> temp.data,
                      tf_obj.tables[i].x_bounds[0],
                      tf_obj.tables[i].x_bounds[1],
                      tf_obj.field_ids[i], tf_obj.weight_field_ids[i],
                      tf_obj.weight_table_ids[i],
                      tf_obj.tables[i].pass_through)
            self.my_field_tables.append((tf_obj.tables[i],
                                         tf_obj.tables[i].y))
            self.field_tables[i].field_id = tf_obj.field_ids[i]
            self.field_tables[i].weight_field_id = tf_obj.weight_field_ids[i]
            print "Field table", i, "corresponds to",
            print self.field_tables[i].field_id,
            print "(Weighted with ", self.field_tables[i].weight_field_id,
            print ")"

        for i in range(6):
            self.field_table_ids[i] = tf_obj.field_table_ids[i]
            print "Channel", i, "corresponds to", self.field_table_ids[i]
            
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef void eval_transfer(self, np.float64_t dt, np.float64_t *dvs,
                                  np.float64_t *rgba, np.float64_t *grad):
        cdef int i, fid, use
        cdef np.float64_t ta, tf, trgba[6], dot_prod
        # NOTE: We now disable this.  I have left it to ease the process of
        # potentially, one day, re-including it.
        #use = 0
        #for i in range(self.n_field_tables):
        #    fid = self.field_tables[i].field_id
        #    if (dvs[fid] >= self.field_tables[i].bounds[0]) and \
        #       (dvs[fid] <= self.field_tables[i].bounds[1]):
        #        use = 1
        #        break
        for i in range(self.n_field_tables):
            self.istorage[i] = FIT_get_value(&self.field_tables[i], dvs)
        # We have to do this after the interpolation
        for i in range(self.n_field_tables):
            fid = self.field_tables[i].weight_table_id
            if fid != -1: self.istorage[i] *= self.istorage[fid]
        for i in range(6):
            trgba[i] = self.istorage[self.field_table_ids[i]]
            #print i, trgba[i],
        #print
        # A few words on opacity.  We're going to be integrating equation 1.23
        # from Rybicki & Lightman.  dI_\nu / ds = -\alpha_\nu I_\nu + j_\nu
        # \alpha_nu = \kappa \rho , but we leave that up to the input
        # transfer function.
        # SOoooooOOOooo, the upshot is that we are doing a rectangular
        # integration here:
        #   I_{i+1} = ds * C_i + (1.0 - ds*alpha_i) * I_i
        for i in range(3):
            # This is the new way: alpha corresponds to opacity of a given
            # slice.  Previously it was ill-defined, but represented some
            # measure of emissivity.
            ta = fmax((1.0 - dt*trgba[i+3]), 0.0)
            rgba[i  ] = dt*trgba[i  ] + ta * rgba[i  ]
            #rgba[i+3] = dt*trgba[i+3] + ta * rgba[i+3]
            # This is the old way:
            #rgba[i  ] += trgba[i] * (1.0 - rgba[i+3])*dt*trgba[i+3]
            #rgba[i+3] += trgba[i] * (1.0 - rgba[i+3])*dt*trgba[i+3]

cdef class VectorPlane:
    cdef public object avp_pos, avp_dir, acenter, aimage
    cdef np.float64_t *vp_pos, *vp_dir, *center, *image,
    cdef np.float64_t pdx, pdy, bounds[4]
    cdef int nv[2]
    cdef int vp_strides[3]
    cdef int im_strides[3]
    cdef int vd_strides[3]
    cdef public object ax_vec, ay_vec
    cdef np.float64_t *x_vec, *y_vec

    def __cinit__(self, 
                  np.ndarray[np.float64_t, ndim=3] vp_pos,
                  np.ndarray vp_dir,
                  np.ndarray[np.float64_t, ndim=1] center,
                  bounds,
                  np.ndarray[np.float64_t, ndim=3] image,
                  np.ndarray[np.float64_t, ndim=1] x_vec,
                  np.ndarray[np.float64_t, ndim=1] y_vec):
        cdef int i, j
        self.avp_pos = vp_pos
        self.avp_dir = vp_dir
        self.acenter = center
        self.aimage = image
        self.ax_vec = x_vec
        self.ay_vec = y_vec
        self.vp_pos = <np.float64_t *> vp_pos.data
        self.vp_dir = <np.float64_t *> vp_dir.data
        self.center = <np.float64_t *> center.data
        self.image = <np.float64_t *> image.data
        self.x_vec = <np.float64_t *> x_vec.data
        self.y_vec = <np.float64_t *> y_vec.data
        self.nv[0] = vp_pos.shape[0]
        self.nv[1] = vp_pos.shape[1]
        for i in range(4): self.bounds[i] = bounds[i]
        self.pdx = (self.bounds[1] - self.bounds[0])/self.nv[0]
        self.pdy = (self.bounds[3] - self.bounds[2])/self.nv[1]
        for i in range(3):
            self.vp_strides[i] = vp_pos.strides[i] / 8
            self.im_strides[i] = image.strides[i] / 8
        if vp_dir.ndim > 1:
            for i in range(3):
                self.vd_strides[i] = vp_dir.strides[i] / 8
        else:
            self.vd_strides[0] = self.vd_strides[1] = self.vd_strides[2] = -1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef void get_start_stop(self, np.float64_t *ex, int *rv):
        # Extrema need to be re-centered
        cdef np.float64_t cx, cy
        cdef int i
        cx = cy = 0.0
        for i in range(3):
            cx += self.center[i] * self.x_vec[i]
            cy += self.center[i] * self.y_vec[i]
        rv[0] = lrint((ex[0] - cx - self.bounds[0])/self.pdx)
        rv[1] = rv[0] + lrint((ex[1] - ex[0])/self.pdx)
        rv[2] = lrint((ex[2] - cy - self.bounds[2])/self.pdy)
        rv[3] = rv[2] + lrint((ex[3] - ex[2])/self.pdy)

    cdef inline void copy_into(self, np.float64_t *fv, np.float64_t *tv,
                        int i, int j, int nk, int strides[3]):
        # We know the first two dimensions of our from-vector, and our
        # to-vector is flat and 'ni' long
        cdef int k
        cdef int offset = strides[0] * i + strides[1] * j
        for k in range(nk):
            tv[k] = fv[offset + k]

    cdef inline void copy_back(self, np.float64_t *fv, np.float64_t *tv,
                        int i, int j, int nk, int strides[3]):
        cdef int k
        cdef int offset = strides[0] * i + strides[1] * j
        for k in range(nk):
            tv[offset + k] = fv[k]

cdef class PartitionedGrid:
    cdef public object my_data
    cdef public object LeftEdge
    cdef public object RightEdge
    cdef np.float64_t *data[6]
    cdef np.float64_t dvs[6]
    cdef np.float64_t left_edge[3]
    cdef np.float64_t right_edge[3]
    cdef np.float64_t dds[3]
    cdef np.float64_t idds[3]
    cdef int dims[3]
    cdef public int parent_grid_id
    cdef public int n_fields

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def __cinit__(self,
                  int parent_grid_id, int n_fields, data,
                  np.ndarray[np.float64_t, ndim=1] left_edge,
                  np.ndarray[np.float64_t, ndim=1] right_edge,
                  np.ndarray[np.int64_t, ndim=1] dims):
        # The data is likely brought in via a slice, so we copy it
        cdef int i, j, k, size
        cdef np.ndarray[np.float64_t, ndim=3] tdata
        self.parent_grid_id = parent_grid_id
        self.LeftEdge = left_edge
        self.RightEdge = right_edge
        for i in range(3):
            self.left_edge[i] = left_edge[i]
            self.right_edge[i] = right_edge[i]
            self.dims[i] = dims[i]
            self.dds[i] = (self.right_edge[i] - self.left_edge[i])/dims[i]
            self.idds[i] = 1.0/self.dds[i]
        self.my_data = data
        self.n_fields = n_fields
        for i in range(n_fields):
            tdata = data[i]
            self.data[i] = <np.float64_t *> tdata.data

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def cast_plane(self, TransferFunctionProxy tf, VectorPlane vp):
        # This routine will iterate over all of the vectors and cast each in
        # turn.  Might benefit from a more sophisticated intersection check,
        # like http://courses.csusm.edu/cs697exz/ray_box.htm
        cdef int vi, vj, hit, i, ni, nj, nn
        cdef int iter[4]
        cdef np.float64_t v_pos[3], v_dir[3], rgba[6], extrema[4]
        hit = 0
        self.calculate_extent(vp, extrema)
        vp.get_start_stop(extrema, iter)
        iter[0] = iclip(iter[0], 0, vp.nv[0])
        iter[1] = iclip(iter[1], 0, vp.nv[0])
        iter[2] = iclip(iter[2], 0, vp.nv[1])
        iter[3] = iclip(iter[3], 0, vp.nv[1])
        if vp.vd_strides[0] == -1:
            for vi in range(iter[0], iter[1]):
                for vj in range(iter[2], iter[3]):
                    vp.copy_into(vp.vp_pos, v_pos, vi, vj, 3, vp.vp_strides)
                    vp.copy_into(vp.image, rgba, vi, vj, 3, vp.im_strides)
                    self.integrate_ray(v_pos, vp.vp_dir, rgba, tf)
                    vp.copy_back(rgba, vp.image, vi, vj, 3, vp.im_strides)
        else:
            # If we do not have an orthographic projection, we have to cast all
            # our rays (until we can get an extrema calculation...)
            for vi in range(vp.nv[0]):
                for vj in range(vp.nv[1]):
                    vp.copy_into(vp.vp_pos, v_pos, vi, vj, 3, vp.vp_strides)
                    vp.copy_into(vp.image, rgba, vi, vj, 3, vp.im_strides)
                    vp.copy_into(vp.vp_dir, v_dir, vi, vj, 3, vp.vd_strides)
                    self.integrate_ray(v_pos, v_dir, rgba, tf)
                    vp.copy_back(rgba, vp.image, vi, vj, 3, vp.im_strides)
        return hit

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef void calculate_extent(self, VectorPlane vp,
                               np.float64_t extrema[4]):
        # We do this for all eight corners
        cdef np.float64_t *edges[2], temp
        edges[0] = self.left_edge
        edges[1] = self.right_edge
        extrema[0] = extrema[2] = 1e300; extrema[1] = extrema[3] = -1e300
        cdef int i, j, k
        for i in range(2):
            for j in range(2):
                for k in range(2):
                    # This should rotate it into the vector plane
                    temp  = edges[i][0] * vp.x_vec[0]
                    temp += edges[j][1] * vp.x_vec[1]
                    temp += edges[k][2] * vp.x_vec[2]
                    if temp < extrema[0]: extrema[0] = temp
                    if temp > extrema[1]: extrema[1] = temp
                    temp  = edges[i][0] * vp.y_vec[0]
                    temp += edges[j][1] * vp.y_vec[1]
                    temp += edges[k][2] * vp.y_vec[2]
                    if temp < extrema[2]: extrema[2] = temp
                    if temp > extrema[3]: extrema[3] = temp
        #print extrema[0], extrema[1], extrema[2], extrema[3]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int integrate_ray(self, np.float64_t v_pos[3],
                                 np.float64_t v_dir[3],
                                 np.float64_t rgba[4],
                                 TransferFunctionProxy tf):
        cdef int cur_ind[3], step[3], x, y, i, n, flat_ind, hit, direction
        cdef np.float64_t intersect_t = 1.0
        cdef np.float64_t iv_dir[3]
        cdef np.float64_t intersect[3], tmax[3], tdelta[3]
        cdef np.float64_t enter_t, dist, alpha, dt, exit_t
        cdef np.float64_t tr, tl, temp_x, temp_y, dv
        for i in range(3):
            if (v_dir[i] < 0):
                step[i] = -1
            else:
                step[i] = 1
            x = (i+1) % 3
            y = (i+2) % 3
            iv_dir[i] = 1.0/v_dir[0]
            tl = (self.left_edge[i] - v_pos[i])*iv_dir[i]
            temp_x = (v_pos[x] + tl*v_dir[x])
            temp_y = (v_pos[y] + tl*v_dir[y])
            if self.left_edge[x] <= temp_x and temp_x <= self.right_edge[x] and \
               self.left_edge[y] <= temp_y and temp_y <= self.right_edge[y] and \
               0.0 <= tl and tl < intersect_t:
                direction = i
                intersect_t = tl
            tr = (self.right_edge[i] - v_pos[i])*iv_dir[i]
            temp_x = (v_pos[x] + tr*v_dir[x])
            temp_y = (v_pos[y] + tr*v_dir[y])
            if self.left_edge[x] <= temp_x and temp_x <= self.right_edge[x] and \
               self.left_edge[y] <= temp_y and temp_y <= self.right_edge[y] and \
               0.0 <= tr and tr < intersect_t:
                direction = i
                intersect_t = tr
        if self.left_edge[0] <= v_pos[0] and v_pos[0] <= self.right_edge[0] and \
           self.left_edge[1] <= v_pos[1] and v_pos[1] <= self.right_edge[1] and \
           self.left_edge[2] <= v_pos[2] and v_pos[2] <= self.right_edge[2]:
            intersect_t = 0.0
        if not ((0.0 <= intersect_t) and (intersect_t < 1.0)): return 0
        for i in range(3):
            intersect[i] = v_pos[i] + intersect_t * v_dir[i]
            cur_ind[i] = <int> floor((intersect[i] +
                                      step[i]*1e-8*self.dds[i] -
                                      self.left_edge[i])*self.idds[i])
            tmax[i] = (((cur_ind[i]+step[i])*self.dds[i])+
                        self.left_edge[i]-v_pos[i])*iv_dir[i]
            # This deals with the asymmetry in having our indices refer to the
            # left edge of a cell, but the right edge of the brick being one
            # extra zone out.
            if cur_ind[i] == self.dims[i] and step[i] < 0:
                cur_ind[i] = self.dims[i] - 1
            if cur_ind[i] < 0 or cur_ind[i] >= self.dims[i]: return 0
            if step[i] > 0:
                tmax[i] = (((cur_ind[i]+1)*self.dds[i])
                            +self.left_edge[i]-v_pos[i])*iv_dir[i]
            if step[i] < 0:
                tmax[i] = (((cur_ind[i]+0)*self.dds[i])
                            +self.left_edge[i]-v_pos[i])*iv_dir[i]
            tdelta[i] = (self.dds[i]*iv_dir[i])
            if tdelta[i] < 0: tdelta[i] *= -1
        # We have to jumpstart our calculation
        enter_t = intersect_t
        while 1:
            # dims here is one less than the dimensions of the data,
            # but we are tracing on the grid, not on the data...
            if (not (0 <= cur_ind[0] < self.dims[0])) or \
               (not (0 <= cur_ind[1] < self.dims[1])) or \
               (not (0 <= cur_ind[2] < self.dims[2])):
                break
            hit += 1
            if tmax[0] < tmax[1]:
                if tmax[0] < tmax[2]:
                    exit_t = fmin(tmax[0], 1.0)
                    self.sample_values(v_pos, v_dir, enter_t, exit_t, cur_ind,
                                       rgba, tf)
                    cur_ind[0] += step[0]
                    enter_t = tmax[0]
                    tmax[0] += tdelta[0]
                else:
                    exit_t = fmin(tmax[2], 1.0)
                    self.sample_values(v_pos, v_dir, enter_t, exit_t, cur_ind,
                                       rgba, tf)
                    cur_ind[2] += step[2]
                    enter_t = tmax[2]
                    tmax[2] += tdelta[2]
            else:
                if tmax[1] < tmax[2]:
                    exit_t = fmin(tmax[1], 1.0)
                    self.sample_values(v_pos, v_dir, enter_t, exit_t, cur_ind,
                                       rgba, tf)
                    cur_ind[1] += step[1]
                    enter_t = tmax[1]
                    tmax[1] += tdelta[1]
                else:
                    exit_t = fmin(tmax[2], 1.0)
                    self.sample_values(v_pos, v_dir, enter_t, exit_t, cur_ind,
                                       rgba, tf)
                    cur_ind[2] += step[2]
                    enter_t = tmax[2]
                    tmax[2] += tdelta[2]
            if enter_t > 1.0: break
        return hit

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef void sample_values(self,
                            np.float64_t v_pos[3],
                            np.float64_t v_dir[3],
                            np.float64_t enter_t,
                            np.float64_t exit_t,
                            int ci[3],
                            np.float64_t *rgba,
                            TransferFunctionProxy tf):
        cdef np.float64_t cp[3], dp[3], temp, dt, t, dv
        cdef np.float64_t grad[3], ds[3]
        grad[0] = grad[1] = grad[2] = 0.0
        cdef int dti, i
        dt = (exit_t - enter_t) / tf.ns # 4 samples should be dt=0.25
        cdef int offset = ci[0] * (self.dims[1] + 1) * (self.dims[2] + 1) \
                        + ci[1] * (self.dims[2] + 1) + ci[2]
        for i in range(3):
            # temp is the left edge of the current cell
            temp = ci[i] * self.dds[i] + self.left_edge[i]
            # this gets us dp as the current first sample position
            dp[i] = (enter_t + 0.5 * dt) * v_dir[i] + v_pos[i] - temp
            dp[i] *= self.idds[i]
            ds[i] = v_dir[i] * self.idds[i] * dt
        for dti in range(tf.ns): 
            for i in range(self.n_fields):
                self.dvs[i] = offset_interpolate(self.dims, dp, self.data[i] + offset)
            #if (dv < tf.x_bounds[0]) or (dv > tf.x_bounds[1]):
            #    continue
            for i in range(3):
                dp[i] += ds[i]
            tf.eval_transfer(dt, self.dvs, rgba, grad)

cdef class GridFace:
    cdef int direction
    cdef public np.float64_t coord
    cdef np.float64_t left_edge[3]
    cdef np.float64_t right_edge[3]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def __init__(self, grid, int direction, int left):
        self.direction = direction
        if left == 1:
            self.coord = grid.LeftEdge[direction]
        else:
            self.coord = grid.RightEdge[direction]
        cdef int i
        for i in range(3):
            self.left_edge[i] = grid.LeftEdge[i]
            self.right_edge[i] = grid.RightEdge[i]
        self.left_edge[direction] = self.right_edge[direction] = self.coord

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int proj_overlap(self, np.float64_t *left_edge, np.float64_t *right_edge):
        cdef int xax, yax
        xax = (self.direction + 1) % 3
        yax = (self.direction + 2) % 3
        if left_edge[xax] >= self.right_edge[xax]: return 0
        if right_edge[xax] <= self.left_edge[xax]: return 0
        if left_edge[yax] >= self.right_edge[yax]: return 0
        if right_edge[yax] <= self.left_edge[yax]: return 0
        return 1

cdef class ProtoPrism:
    cdef np.float64_t left_edge[3]
    cdef np.float64_t right_edge[3]
    cdef public object LeftEdge
    cdef public object RightEdge
    cdef public object subgrid_faces
    cdef public int parent_grid_id
    def __cinit__(self, int parent_grid_id,
                  np.ndarray[np.float64_t, ndim=1] left_edge,
                  np.ndarray[np.float64_t, ndim=1] right_edge,
                  subgrid_faces):
        self.parent_grid_id = parent_grid_id
        cdef int i
        self.LeftEdge = left_edge
        self.RightEdge = right_edge
        for i in range(3):
            self.left_edge[i] = left_edge[i]
            self.right_edge[i] = right_edge[i]
        self.subgrid_faces = subgrid_faces

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def sweep(self, int direction = 0, int stack = 0):
        cdef int i
        cdef GridFace face
        cdef np.float64_t proto_split[3]
        for i in range(3): proto_split[i] = self.right_edge[i]
        for face in self.subgrid_faces[direction]:
            proto_split[direction] = face.coord
            if proto_split[direction] <= self.left_edge[direction]:
                continue
            if proto_split[direction] == self.right_edge[direction]:
                if stack == 2: return [self]
                return self.sweep((direction + 1) % 3, stack + 1)
            if face.proj_overlap(self.left_edge, proto_split) == 1:
                left, right = self.split(proto_split, direction)
                LC = left.sweep((direction + 1) % 3)
                RC = right.sweep(direction)
                return LC + RC
        raise RuntimeError

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef object split(self, np.float64_t *sp, int direction):
        cdef int i
        cdef np.ndarray split_left = self.LeftEdge.copy()
        cdef np.ndarray split_right = self.RightEdge.copy()

        for i in range(3): split_left[i] = self.right_edge[i]
        split_left[direction] = sp[direction]
        left = ProtoPrism(self.parent_grid_id, self.LeftEdge, split_left,
                          self.subgrid_faces)

        for i in range(3): split_right[i] = self.left_edge[i]
        split_right[direction] = sp[direction]
        right = ProtoPrism(self.parent_grid_id, split_right, self.RightEdge,
                           self.subgrid_faces)

        return (left, right)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def get_brick(self, np.ndarray[np.float64_t, ndim=1] grid_left_edge,
                        np.ndarray[np.float64_t, ndim=1] grid_dds,
                        child_mask):
        # We get passed in the left edge, the dds (which gives dimensions) and
        # the data, which is already vertex-centered.
        cdef PartitionedGrid PG
        cdef int li[3], ri[3], idims[3], i
        for i in range(3):
            li[i] = lrint((self.left_edge[i] - grid_left_edge[i])/grid_dds[i])
            ri[i] = lrint((self.right_edge[i] - grid_left_edge[i])/grid_dds[i])
            idims[i] = ri[i] - li[i]
        if child_mask[li[0], li[1], li[2]] == 0: return []
        cdef np.ndarray[np.int64_t, ndim=1] dims = np.empty(3, dtype='int64')
        for i in range(3):
            dims[i] = idims[i]
        #cdef np.ndarray[np.float64_t, ndim=3] new_data
        #new_data = data[li[0]:ri[0]+1,li[1]:ri[1]+1,li[2]:ri[2]+1].copy()
        #PG = PartitionedGrid(self.parent_grid_id, new_data,
        #                     self.LeftEdge, self.RightEdge, dims)
        return ((li[0], ri[0]), (li[1], ri[1]), (li[2], ri[2]), dims)



"""
Various definitions for various other modules and routines

@author: U{JS Oishi<http://www.jsoishi.org/>}
@organization: U{UC Berkeley<http://www.astro.berkeley.edu/>}
@contact: U{jsoishi@astro.berkeley.edu<mailto:jsoishi@astro.berkeley.edu>}

@todo: Move into yt.Defs, along with enki.EnkiDefs
@license:
  Copyright (C) 2008-2009 J.S. Oishi.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
from yt.funcs import *
### this assumes EnzoDefs.py has *already* been imported

# converts the Orion inputs file name to the Enzo/yt name expected
# throughout the code. key is Orion name, value is Enzo/yt equivalent
orion2enzoDict = {"amr.n_cell": "TopGridRank",
                  "materials.gamma": "Gamma",
                  "amr.ref_ratio": "RefineBy"
                  }

yt2orionFieldsDict = {}
orion2ytFieldsDict = {}

orion_FAB_header_pattern = r"^FAB \(\((\d+), \([0-9 ]+\)\),\(\d+, \(([0-9 ]+)\)\)\)\(\((\d+,\d+,\d+)\) \((\d+,\d+,\d+)\) \((\d+,\d+,\d+)\)\) (\d+)\n"



from Forthon import *
from fKDpy import *
import numpy,random

n = 32768


fKD.tags = fzeros((64),'i')
fKD.dist = fzeros((64),'d')
fKD.pos = fzeros((3,n),'d')
fKD.nn = 64
fKD.nparts = n
fKD.sort = True
fKD.rearrange = True
fKD.qv = numpy.array([16./32, 16./32, 16./32])

fp = open('parts.txt','r')
xpos = []
ypos = []
zpos = []
line = fp.readline()
while line:
    line = line.split()
    xpos.append(float(line[0]))
    ypos.append(float(line[1]))
    zpos.append(float(line[2]))
    line= fp.readline()

fp.close()


for k in range(32):
    for j in range(32):
        for i in range(32):
            fKD.pos[0][i + j*32 + k*1024] = float(i)/32 + 1./64 + 0.0001*random.random()
            fKD.pos[1][i + j*32 + k*1024] = float(j)/32 + 1./64 + 0.0001*random.random()
            fKD.pos[2][i + j*32 + k*1024] = float(k)/32 + 1./64 + 0.0001*random.random()

            

#print fKD.pos[0][0],fKD.pos[1][0],fKD.pos[2][0]

create_tree()


find_nn_nearest_neighbors()

#print 'next'

#fKD.qv = numpy.array([0., 0., 0.])

#find_nn_nearest_neighbors()


#print (fKD.tags - 1)
#print fKD.dist

free_tree()



"""
Logging facilities for Reason

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2008-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.reason import *

import logging, sys
import wx.lib.mixins.listctrl as listmix

class LoggingListBox(wx.ListCtrl, listmix.ListCtrlAutoWidthMixin):
    def __init__(self, parent, dataSource):
        wx.ListCtrl.__init__(self, parent,
                             style=wx.LC_REPORT|wx.LC_SINGLE_SEL|wx.LC_VIRTUAL)
        listmix.ListCtrlAutoWidthMixin.__init__(self)
        columns = dataSource.GetColumnHeaders()
        for col, text in enumerate(columns):
            self.InsertColumn(col, text)
        self.SetItemCount(dataSource.GetCount())
        self.dataSource = dataSource

    def UpdateCounts(self):
        c = self.dataSource.GetCount()
        self.SetItemCount(c)
        self.EnsureVisible(c-1)
        self.Refresh()

    def OnGetItemText(self, item, col):
        return self.dataSource.GetItem(item)[col]

    def OnGetItemAttr(self, item): return None
    def OnGetItemImage(self, item): return -1

class LoggingWindowBox(wx.Panel):
    def __init__(self, parent, id):
        wx.Panel.__init__(self, parent, id)
        self.__redirect_logging()
        self.__setup_controls()
        mylog.info("Welcome to Reason.")

    def __setup_controls(self):
        self.event_list = LoggingListBox(self, self.handler.output)
        self.handler.setTextBox(self.event_list)
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.event_list, 1, wx.EXPAND)
        self.SetSizer(self.sizer)
        self.Fit()

    def __redirect_logging(self):
        yt.logger.disable_stream_logging()
        self.handler = OutputWindowStream()
        self.handler.setFormatter(logging.Formatter(yt.logger.fstring))
        yt.logger.rootLogger.addHandler(self.handler)

class OutputWindowStream(logging.Handler):
    # Shamelessly borrowed from:
    #   http://lists.wxwidgets.org/pipermail/wxpython-users/2007-October/069288.html
    def __init__(self):
        logging.Handler.__init__(self)
        self.output = LoggingDataSource()
        self.buffer = []
        self.events = []

    def setTextBox(self, textbox):
        """
        textbox presents the instance of textctrl in the mainframe
        Method is called after frame is up and running
        """
        self.textbox = textbox

    def shouldFlush(self):
        if self.textbox:
            return True
        else:
            return False

    def emit(self, record):
        self.buffer.append(record)
        if self.shouldFlush():
            self.flush()

    def flush(self):
        """
        Thread safe flush method which schedule the
        textbox update after a wx main loop ends
        """
        for record in self.buffer:
            self.format(record)
            wx.CallAfter(self.output.AppendItem, record)
            wx.CallAfter(self.textbox.UpdateCounts)
            wx.CallAfter(wx.YieldIfNeeded)
        self.buffer = []

    def close(self):
        self.flush()
        self.textbox = None
        logging.Handler.close(self)

class LoggingDataSource(object):
    cols = ("Component","Level","Time","Message")
    attrs = ("name","levelname","asctime","message")
    def __init__(self):
        self.records = []

    def GetColumnHeaders(self):
        return self.cols

    def GetCount(self):
        return len(self.records)

    def GetItem(self, index):
        return [getattr(self.records[index],att) for att in self.attrs]

    def UpdateCache(self, start, end):
        pass

    def AppendItem(self, rec):
        self.records.append(rec)



# Merger.py. Written by Stephen Skory & Rick Wagner, August 2006.

# figures the 'family tree,' and outputs in GraphViz format, the galaxy
# merger history for a series of files that list the particles & the group that
# they are assigned to.

# this file only contains functions that are used by the output of fastBuildMerge.py.

# June 2008, converted/upgraded for use in Matthew Turk's enzo analyzer yt by Stephen Skory

# define the Group class for the list of groups, which contains
# the group id, a list of the particles in that group, the numerical
# order of the group, the percentage of the groups particles that go to the 
# ether (no group), and the percentage of the group that comes from no group.
class Group:


    def __init__(self, id, particles, orderIndex, toEther, fromEther, flag):
        self.id = id
        self.particles = particles
        self.orderIndex = orderIndex
        # these are fields to track how much of each group comes and goes to the
        # ether
        self.toEther = toEther
        self.fromEther = fromEther
        # the flag says whether or not this group is linked to the final,
        # interesting group(s).
        self.flag = flag

# the heavy lifting. This takes in a parentGroup and sees how many of its particles
# go to the childGroup.
# DEPRECIATED, replaced by pushListCount. I've left it here because it's still functional and
# a lot simpler to understand than pushListCount. If you want to use it, uncomment out the line
# where it's used in buildLinks, and comment out the pushListCount line. This is a whole bunch slower
# than pushListCount, so I don't know why you would want to do that.

    def isParent(self, childGroup):

        n = 0
        for particleNumber in self.particles:
            if childGroup.particles.count(particleNumber):
                #return True
                n+=1

        return n

# this function takes two lists which have both been sorted in increasing order. This is a way to
# take the O(n^2) matching and make it roughly O(2n). Since most of the time most particles go from
# one group to the next, a step down one list will get the same next particle in the other list, so
# most of the time we can move down both lists simultaneously. When one entry is bigger than the other,
# we use findNextMatch to see how far to move the second list until it matches the first, or 
# surpasses the first (in which case we then move the first list until it catches up with the second, 
# ad naseum).
def pushListCount(list1, list2):


    count = 0
    i = 0
    j = 0
    
    while (i < len(list1)) and (j < len(list2)):
        #print 'list1['+str(i)+']='+str(list1[i])+' list2['+str(j)+']='+str(list2[j])
        if list1[i] == list2[j]:
            i += 1
            j += 1
            count += 1
            #print 'Matched first try'
        else:
            if (i <= len(list1) -1) and (j <= len(list2) -1):
                value = min(list1[i], list2[j])
                                #print str(value)
                if value == list1[i]:
                    #print 'First list smaller'
                    nextindex = findNextMatch(i+1, list2[j], list1)
                    if nextindex > -1:
                        if nextindex == 1.5: # special case
                            j += 1
                            i += 1
                        else:
                            i = nextindex + 1
                            j += 1
                            count += 1
                    if nextindex < -1:
                        i = -1 * nextindex
                        j += 1
                    if (j == len(list2)) or (i == len(list1)):
                        break
                    if nextindex == -1:
                        #print "Got minus one on second list, nextindex= " + str(nextindex)
                        break

                if value == list2[j]:
                    #print 'Second list smaller'
                    nextindex = findNextMatch(j+1, list1[i], list2)
                    if nextindex > -1:
                        if nextindex == 1.5: # special case
                            j += 1
                            i +=1
                        else:
                            j = nextindex +1
                            i += 1
                            count += 1
                    if nextindex < -1:
                        j = -1 * nextindex
                        i += 1
                    if nextindex == 1:
                        j = nextindex + 1
                        i += 1
                    if (j == len(list2)) or (i == len(list1)):
                        break
                    if nextindex == -1:
                        #print "Got minus one on first list, nextindex= " + str(nextindex)
                        break
        #print 'count: ' + str(count)

    return count

def findNextMatch(startIndex, value, thislist):
    #print "looking for " + str(value) + ', startindex= ' + str(startIndex)
    outval = -1
    for i in range(startIndex, len(thislist)):
        #print 'thislist[' + str(i) +'] = ' +str(thislist[i])
        if thislist[i] == value:
            outval = i
            break
        if thislist[i] > value:
            outval = -1 * i
            if outval == -1: # this in case i=1, which isn't a problem unless outval=-1 *is* the value we want
                         # this happens if the second item in each list matches
                outval = 1.5
            break
    #print 'found match for ' + str(value) + ' at index: ' + str(outval)

    return outval
    


# add yt.hop particles into the Group structure
def convertGroups(snapshot,orderIndex,hardmaxgroup):
    from yt.lagos import mylog
    # ytHopResults is the stuff from yt containing the results of hop.
    # orderIndex is the index for this snapshot
    # hardmaxgroup is some reasonable limit for the number of galaxies looked at for each
    # time step.
    
    groups = []
    particles = []
    # set up the percent fields for later...
    toEther = 100
    fromEther = 100
    pcount = 0 # counts how many particles are groups for that particular time step
    flag = 0 # zero means not assoc with final group, one means yes.
    
    # loop over the groups
    for i,group in enumerate(snapshot):
        # don't go past hardmaxgroup
        if (i >= hardmaxgroup):
            break
        # add it to our groups as a Group
        groups.append(Group(i,group["particle_index"],orderIndex,toEther,fromEther,flag))
        pcount += len(group["particle_index"])
    
    string = "There are " + str(pcount) + " particles in orderIndex " + str(orderIndex) + "."
    mylog.info(string)
    return(groups)


# from the yt.hop output, add to the positions dict the central position and group index for this
# snapshot
def convertPositions(snapshot,orderIndex,positions,hardmaxgroup):
    from yt.lagos import mylog
    # loop over the hop groups
    for i,group in enumerate(snapshot):
        # if we're trying to read in too many groups, we're done so we leave
        if (i >= hardmaxgroup):
            break
        center = group.maximum_density_location()
        positions[orderIndex,i] = (center[0],center[1],center[2])
    
    string = 'There are ' + str(len(snapshot)) + ' groups in orderIndex ' + str(orderIndex)
    mylog.info(string)
    
    return (positions)

# add yt.hop particles into the Group structure, but for selected groups, for the 
# last time step, which is the first inspected.
def convertGroupsSelected(snapshot,orderIndex,indices):
    from yt.lagos import mylog
    # ytHopResults is the stuff from yt containing the results of hop.
    # orderIndex is the index for this snapshot
    # hardmaxgroup is some reasonable limit for the number of galaxies looked at for each
    # time step.
    
    groups = []
    particles = []
    # set up the percent fields for later...
    toEther = 100
    fromEther = 100
    pcount = 0 # counts how many particles are groups for that particular time step
    flag = 0 # zero means not assoc with final group, one means yes.
    
    # loop over the selected groups
    for index in indices:
        # pick out the one group
        group = snapshot[index]
        
        # add it to our groups as a Group
        groups.append(Group(index,group["particle_index"],orderIndex,toEther,fromEther,flag))
        pcount += len(group["particle_index"])
    
    string = "There are " + str(pcount) + " particles in orderIndex " + str(orderIndex) + "."
    mylog.info(string)
    return(groups)

# from the yt.hop output, add to the positions dict the central position and group index for this
# snapshot, but only chosen groups, which is what we want at the final data dump
def convertPositionsSelected(snapshot,orderIndex,positions,indices):
    from yt.lagos import mylog
    
    # loop over the selected groups
    for index in indices:
        group = snapshot[index]
        center = group.maximum_density_location()
        positions[orderIndex,index] = (center[0],center[1],center[2])
    
    string = 'There are ' + str(len(indices)) +' groups in orderIndex ' + str(orderIndex)
    mylog.info(string)
    
    return (positions)



# this does three things:
# 1. Builds the links list, which contains the parentGroup.id and childGroup.id
# of the matched pair, along with the number of particles that are transferred.
# 2. Calculates what percentage of the parentGroup is transferred to the child
# groups. This is then subtracted from 100 to give how much of the group goes to
# the ether.
# 3. Similarly, calculates what percentage of the child group comes from the
# ether.
def buildLinks(parentGroups, childGroups,positions):
    from yt.lagos import mylog
    links = []
    childPercents = {}
    
    for parentGroup in parentGroups:
        parentPercent = 0
        for childGroup in childGroups:
            number = 0 # this will be the number of particles that go from the parent
            # group to the child group.
            # Below we calculate the distance between the groups, taking the periodic
            # boundary conditions into account.
            dist_x = abs(positions[parentGroup.orderIndex,parentGroup.id][0] - positions[childGroup.orderIndex,childGroup.id][0])
            if dist_x > 0.5:
                dist_x = 1 - dist_x
            dist_y = abs(positions[parentGroup.orderIndex,parentGroup.id][1] - positions[childGroup.orderIndex,childGroup.id][1])
            if dist_y > 0.5:
                dist_y = 1 - dist_y
            dist_z = abs(positions[parentGroup.orderIndex,parentGroup.id][2] - positions[childGroup.orderIndex,childGroup.id][2])
            if dist_z > 0.5:
                dist_z = 1 - dist_z
            dist = (dist_x**2 + dist_y**2 + dist_z**2)**0.5
            # If the childgroup isn't flagged, or if the two groups are too far away,
            # the groups don't get checked; number stays equal to zero; and
            # we move to the next parentgroup.
            if (childGroup.flag == 1) and (dist <= 0.1):
                string = 'checking parentGroup (' + str(len(parentGroup.particles)) + '): ' + str(parentGroup.orderIndex) + '.' + str(parentGroup.id) + \
                ' with childGroup (' + str(len(childGroup.particles)) + '): ' + str(childGroup.orderIndex) + '.' + str(childGroup.id)
                mylog.info(string)
                #number = parentGroup.isParent(childGroup)
                number = pushListCount(parentGroup.particles,childGroup.particles)
                #string = 'number: ' + str(number)
                #mylog.info(string)
            # create an entry in the childPercents dictionary if it's not already
            # there
            if childPercents.has_key(childGroup.id) != 1:
                childPercents[childGroup.id] = 0
            
            # If the number of particles from the parentgroup goes to the child group,
            # above some threshold, add this link to the links list; calculate how much
            # of the parent group went to the child group and add that to any existing total
            # of particles leaving the parent group; add the particles going to the child
            # group to it's percentage of particles not coming from the ether; and finally
            # flag the parentgroup as part of the lineage for future checking.
            if (number/float(len(parentGroup.particles)))*100 > 50:
                links.append([parentGroup.id,childGroup.id,number])
                string = 'parentGroup: ' + str(parentGroup.orderIndex) + '.' + str(parentGroup.id) + \
                ' | childGroup: ' + str(childGroup.orderIndex) + '.' + str(childGroup.id) + \
                ' | %% transferred: %2.2f%%' % (number/float(len(parentGroup.particles)) * 100)
                mylog.info(string)
                parentPercent += (number/float(len(parentGroup.particles))*100)
                childPercents[childGroup.id] += (number/float(len(childGroup.particles))*100)
                parentGroup.flag = 1
        # End loop over childGroups.
        # if any of the particles from the parent group went to a child group,
        # change the parent groups toEther to reflect this.
        if parentPercent > 0:
            parentPercent = 100 - parentPercent
            parentGroup.toEther = parentPercent

    # go over the childgroups and see how much of the particles came from
    # the ether.
    for childGroup in childGroups:
        if childPercents[childGroup.id] > 0:
            childPercents[childGroup.id] = 100 - childPercents[childGroup.id]
            childGroup.fromEther = childPercents[childGroup.id]

    return (links,parentGroups,childGroups)

# The functions below just write out simple blocks that GraphViz needs.
def writeTop(fopen):

    line = 'digraph galaxy {size="40,40";\n'
    fopen.write(line)

def writeClose(fopen):

    line = '};\n'
    fopen.write(line)

def writeOpen(fopen):

    line = '{\n'
    fopen.write(line)

def writeNode(fopen):
    
    line = 'node [color=lightblue, style=bold, shape=record];\n'
    fopen.write(line)

# In order to keep the boxes for the haloes on the same level, GraphViz needs
# to be told which boxes are all on the same level.
# Mar 2007, added lines for redshift levels.
def writeLevels(fopen, Groups, redshift):

    line = '{ rank = same;\n'
    fopen.write(line)

    line = '"' + str(redshift) + '";'
    fopen.write(line)
    
    for Group in Groups:
        if (Group.toEther != 100) or (Group.fromEther != 100):
            line = '"' + str(Group.orderIndex) + '.' + str(Group.id) + '";'
            fopen.write(line)
    
    line = "\n};\n"
    
    fopen.write(line)

    
# Each box (node) has four bits of information:
# 1. For all but the top nodes, the percentage of the particles that come from
# the ether.
# 2. The number of particles in it.
# 3. The center of that group in (0,0,0)->(1,1,1) coordinates
# 4. The all but the bottom notes, the percentage of the particles that go to 
# the ether.

def writeLabels(fopen, Groups, gPositions, switch):

    # switch = 0: very top node, beginning of time
    # switch = 1: intermediate node, many
    # switch = 2: very bottom node, singular (usually), the last in time

    groupLength = len(Groups)
    
    if switch == 0:
        for Group in Groups:
            if (Group.toEther != 100) or (Group.fromEther != 100):
                color = 1 - float(Group.id)/float(groupLength)
                line = '"' + str(Group.orderIndex) + '.' + str(Group.id) + '" [label="{' + str(len(Group.particles))
                # write the position triplet for the group
                line += '\\n(%1.3f,' % float(gPositions[Group.orderIndex,Group.id][0])
                line += '%1.3f,' % float(gPositions[Group.orderIndex,Group.id][1])
                line += '%1.3f)' % float(gPositions[Group.orderIndex,Group.id][2])
                line += '| %2.2f%%}", shape="record",' % Group.toEther
                line += 'color="%1.3f ' % color
                line += '1.0 %1.3f"];\n' % color
                fopen.write(line)
        
    if switch == 1:
        for Group in Groups:
            if (Group.toEther != 100) or (Group.fromEther != 100):
                color = 1 - float(Group.id)/float(groupLength)
                line = '"' + str(Group.orderIndex) + '.' + str(Group.id)
                line += '" [label="{%2.2f%%| {' % Group.fromEther
                line += str(len(Group.particles))
                # write the position triplet for the group
                line += '\\n(%1.3f,' % float(gPositions[Group.orderIndex,Group.id][0])
                line += '%1.3f,' % float(gPositions[Group.orderIndex,Group.id][1])
                line += '%1.3f)' % float(gPositions[Group.orderIndex,Group.id][2])
                line +='}|%2.2f%%}", shape="record"' % Group.toEther
                line += ', color="%1.3f ' % color
                line += '1.0 %1.3f"];\n' % color
                fopen.write(line)
            
    if switch == 2:
        for Group in Groups:
            if (Group.toEther != 100) or (Group.fromEther != 100):
                color = 1 - float(Group.id)/float(groupLength)
                line = '"' + str(Group.orderIndex) + '.' + str(Group.id)
                line += '" [label="{%2.2f%%|' % Group.fromEther
                line += str(len(Group.particles))
                # write the position triplet for the group
                line += '\\n(%1.3f,' % float(gPositions[Group.orderIndex,Group.id][0])
                line += '%1.3f,' % float(gPositions[Group.orderIndex,Group.id][1])
                line += '%1.3f)' % float(gPositions[Group.orderIndex,Group.id][2])
                line += '}", shape="record", color="%1.3f ' % color
                line += '1.0 %1.3f"];\n' % color
                fopen.write(line)

# write out the GraphViz links that connect two groups.
def writeLinks(fopen, links, parentGroups, childGroups):

    for parentGroup in parentGroups:
        for link in links:
            if str(parentGroup.id) == str(link[0]):
                line = '"' + str(parentGroup.orderIndex) + '.' + str(link[0]) + '"->"' + str(parentGroup.orderIndex+1) + '.' + str(link[1])+\
                '"[label="%2.2f%%",color="blue", fontsize=10];\n' % (link[2]/float(len(parentGroup.particles)) * 100)
                fopen.write(line)



"""
Commonly used mathematical functions.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: UCSD Physics/CASS
Author: Stephen Skory <stephenskory@yahoo.com>
Affiliation: UCSD Physics/CASS
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2008-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as na
import math

def periodic_dist(a, b, period):
    """
    Find the Euclidian periodic distance between two points.
    *a* and *b* are array-like vectors, and *period* is a float or
    array-like value for the periodic size of the volume.
    """
    a = na.array(a)
    b = na.array(b)
    if a.size != b.size: RunTimeError("Arrays must be the same shape.")
    c = na.empty((2, a.size), dtype="float64")
    c[0,:] = abs(a - b)
    c[1,:] = period - abs(a - b)
    d = na.amin(c, axis=0)**2
    return math.sqrt(d.sum())

    


"""
Import the components of the volume rendering extension

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as na
from grid_partitioner import HomogenizedVolume
from transfer_functions import ProjectionTransferFunction
from yt.funcs import *
import yt.amr_utils as au

class Camera(object):
    def __init__(self, center, normal_vector, width,
                 resolution, transfer_function,
                 north_vector = None,
                 volume = None, fields = None,
                 log_fields = None,
                 sub_samples = 5, pf = None):
        r"""A viewpoint into a volume, for volume rendering.

        The camera represents the eye of an observer, which will be used to
        generate ray-cast volume renderings of the domain.

        Parameters
        ----------
        center : array_like
            The current "center" of the view port -- the focal point for the
            camera.
        normal_vector : array_like
            The vector between the camera position and the center.
        width : float or list of floats
            The current width of the image.  If a single float, the volume is
            cubical, but if not, it is front/back, left/right, top/bottom.
        resolution : int or list of ints
            The number of pixels in each direction.
        north_vector : array_like, optional
            The "up" direction for the plane of rays.  If not specific, calculated
            automatically.
        volume : `yt.extensions.volume_rendering.HomogenizedVolume`, optional
            The volume to ray cast through.  Can be specified for finer-grained
            control, but otherwise will be automatically generated.
        fields : list of fields, optional
            This is the list of fields we want to volume render; defaults to
            Density.
        log_fields : list of bool, optional
            Whether we should take the log of the fields before supplying them to
            the volume rendering mechanism.
        sub_samples : int, optional
            The number of samples to take inside every cell per ray.
        pf : `~yt.lagos.StaticOutput`
            For now, this is a require parameter!  But in the future it will become
            optional.  This is the parameter file to volume render.

        Examples
        --------

        >>> cam = vr.Camera(c, L, W, (N,N), transfer_function = tf, pf = pf)
        >>> image = cam.snapshot()
        """
        if pf is not None: self.pf = pf
        if not iterable(resolution):
            resolution = (resolution, resolution)
        self.resolution = resolution
        self.sub_samples = sub_samples
        if not iterable(width):
            width = (width, width, width) # front/back, left/right, top/bottom
        self.width = width
        self.center = center
        if fields is None: fields = ["Density"]
        self.fields = fields
        if transfer_function is None:
            transfer_function = ProjectionTransferFunction()
        self.transfer_function = transfer_function
        self._setup_normalized_vectors(normal_vector, north_vector)
        self.log_fields = log_fields
        if volume is None:
            volume = HomogenizedVolume(fields, pf = self.pf,
                                       log_fields = log_fields)
        self.volume = volume

    def _setup_normalized_vectors(self, normal_vector, north_vector):
        # Now we set up our various vectors
        normal_vector /= na.sqrt( na.dot(normal_vector, normal_vector))
        if north_vector is None:
            vecs = na.identity(3)
            t = na.cross(normal_vector, vecs).sum(axis=1)
            ax = t.argmax()
            north_vector = na.cross(vecs[ax,:], normal_vector).ravel()
        north_vector /= na.sqrt(na.dot(north_vector, north_vector))
        east_vector = -na.cross(north_vector, normal_vector).ravel()
        east_vector /= na.sqrt(na.dot(east_vector, east_vector))
        self.unit_vectors = [north_vector, east_vector, normal_vector]
        self.box_vectors = na.array([self.unit_vectors[0]*self.width[0],
                                     self.unit_vectors[1]*self.width[1],
                                     self.unit_vectors[2]*self.width[2]])

        self.origin = self.center - 0.5*self.width[0]*self.unit_vectors[0] \
                                  - 0.5*self.width[1]*self.unit_vectors[1] \
                                  - 0.5*self.width[2]*self.unit_vectors[2]
        self.back_center = self.center - 0.5*self.width[0]*self.unit_vectors[2]
        self.front_center = self.center + 0.5*self.width[0]*self.unit_vectors[2]
        self.inv_mat = na.linalg.pinv(self.unit_vectors)

    def look_at(self, new_center, north_vector = None):
        r"""Change the view direction based on a new focal point.

        This will recalculate all the necessary vectors and vector planes related
        to a camera to point at a new location.

        Parameters
        ----------
        new_center : array_like
            The new "center" of the view port -- the focal point for the
            camera.
        north_vector : array_like, optional
            The "up" direction for the plane of rays.  If not specific,
            calculated automatically.
        """
        normal_vector = self.front_center - new_center
        self._setup_normalized_vectors(normal_vector, north_vector)

    def get_vector_plane(self, image):
        # We should move away from pre-generation of vectors like this and into
        # the usage of on-the-fly generation in the VolumeIntegrator module
        # We might have a different width and back_center
        px = na.linspace(-self.width[0]/2.0, self.width[0]/2.0,
                         self.resolution[0])[:,None]
        py = na.linspace(-self.width[1]/2.0, self.width[1]/2.0,
                         self.resolution[1])[None,:]
        inv_mat = self.inv_mat
        bc = self.back_center
        positions = na.zeros((self.resolution[0], self.resolution[1], 3),
                          dtype='float64', order='C')
        positions[:,:,0] = inv_mat[0,0]*px+inv_mat[0,1]*py+self.back_center[0]
        positions[:,:,1] = inv_mat[1,0]*px+inv_mat[1,1]*py+self.back_center[1]
        positions[:,:,2] = inv_mat[2,0]*px+inv_mat[2,1]*py+self.back_center[2]
        bounds = (px.min(), px.max(), py.min(), py.max())
        vector_plane = au.VectorPlane(positions, self.box_vectors[2],
                                      self.back_center, bounds, image,
                                      self.unit_vectors[0],
                                      self.unit_vectors[1])
        return vector_plane

    def snapshot(self):
        r"""Ray-cast the camera.

        This method instructs the camera to take a snapshot -- i.e., call the ray
        caster -- based on its current settings.

        Returns
        -------
        image : array
            An (N,M,3) array of the final returned values, in float64 form.
        """
        image = na.zeros((self.resolution[0], self.resolution[1], 3),
                         dtype='float64', order='C')
        vector_plane = self.get_vector_plane(image)
        tfp = au.TransferFunctionProxy(self.transfer_function) # Reset it every time
        tfp.ns = self.sub_samples
        self.volume.initialize_source()
        pbar = get_pbar("Ray casting",
                        (self.volume.brick_dimensions + 1).prod(axis=-1).sum())
        total_cells = 0
        for brick in self.volume.traverse(self.back_center, self.front_center):
            brick.cast_plane(tfp, vector_plane)
            total_cells += na.prod(brick.my_data[0].shape)
            pbar.update(total_cells)
        pbar.finish()
        return image

    def zoom(self, factor):
        r"""Change the distance to the focal point.

        This will zoom the camera in by some `factor` toward the focal point,
        along the current view direction, modifying the left/right and up/down
        extents as well.

        Parameters
        ----------
        factor : float
            The factor by which to reduce the distance to the focal point.


        Notes
        -----

        You will need to call snapshot() again to get a new image.

        """
        self.width = [w / factor for w in self.width]
        self._setup_normalized_vectors(
                self.unit_vectors[2], self.unit_vectors[0])

    def zoomin(self, final, n_steps):
        r"""Loop over a zoomin and return snapshots along the way.

        This will yield `n_steps` snapshots until the current view has been
        zooming in to a final factor of `final`.

        Parameters
        ----------
        final : float
            The zoom factor, with respect to current, desired at the end of the
            sequence.
        n_steps : int
            The number of zoom snapshots to make.


        Examples
        --------

        >>> for i, snapshot in enumerate(cam.zoomin(100.0, 10)):
        ...     iw.write_bitmap(snapshot, "zoom_%04i.png" % i)
        """
        f = final**(1.0/n_steps)
        for i in xrange(n_steps):
            self.zoom(f)
            yield self.snapshot()


class PerspectiveCamera(Camera):
    def get_vector_plane(self, image):
        # We should move away from pre-generation of vectors like this and into
        # the usage of on-the-fly generation in the VolumeIntegrator module
        # We might have a different width and back_center
        px = na.linspace(-self.width[0]/2.0, self.width[0]/2.0,
                         self.resolution[0])[:,None]
        py = na.linspace(-self.width[1]/2.0, self.width[1]/2.0,
                         self.resolution[1])[None,:]
        inv_mat = self.inv_mat
        bc = self.back_center
        positions = na.zeros((self.resolution[0], self.resolution[1], 3),
                          dtype='float64', order='C')
        positions[:,:,0] = inv_mat[0,0]*px+inv_mat[0,1]*py+self.back_center[0]
        positions[:,:,1] = inv_mat[1,0]*px+inv_mat[1,1]*py+self.back_center[1]
        positions[:,:,2] = inv_mat[2,0]*px+inv_mat[2,1]*py+self.back_center[2]
        bounds = (px.min(), px.max(), py.min(), py.max())
        
        # We are likely adding on an odd cutting condition here
        vectors = self.front_center - positions
        n = vectors * vectors
        n = na.sum(n, axis=2)**0.5
        vectors /= n[:,:,None]
        print vectors.shape, vectors.dtype, vectors.flags

        vector_plane = au.VectorPlane(positions, vectors,
                                      self.back_center, bounds, image,
                                      self.unit_vectors[0],
                                      self.unit_vectors[1])
        return vector_plane

class StereoPairCamera(Camera):
    def __init__(self, original_camera, relative_separation = 0.005):
        self.original_camera = original_camera
        self.relative_separation = relative_separation

    def split(self):
        oc = self.original_camera
        uv = oc.unit_vectors
        c = oc.center
        fc = oc.front_center
        wx, wy, wz = oc.width
        left_normal = fc + uv[1] * 0.5*self.relative_separation * wx - c
        right_normal = fc - uv[1] * 0.5*self.relative_separation * wx - c
        left_camera = Camera(c, left_normal, oc.width,
                             oc.resolution, oc.transfer_function, uv[0],
                             oc.volume, oc.fields, oc.log_fields,
                             oc.sub_samples, oc.pf)
        right_camera = Camera(c, right_normal, oc.width,
                             oc.resolution, oc.transfer_function, uv[0],
                             oc.volume, oc.fields, oc.log_fields,
                             oc.sub_samples, oc.pf)
        return (left_camera, right_camera)



#!/usr/bin/python

# Stephen Skory
# sskory@physics.ucsd.edu
# created August 2006
# modified June 2008
# this python script builds a script for creating halo merger trees using hop inside yt
# see http://yt.enzotools.org/ for more info on yt and this script.

# the output of GraphViz at the very will have boxes connected by arrows.
# The arrow into a box is how much of the previous halo's DM & stars is
# contributing to that halo.
# Similarly the arrow leaving a box is how much of the DM & stars is leaving
# that halo for the next one.
# Inside the box, the top % is how much of the halo is accreted from the 
# 'ether'; particles that in the previous time step were assigned to no group.
# Similarly, the bottom % is how much of the halo leaves and is assigned to no
# group in the next time step.
# The middle of the box shows how many star/DM particles are in the halo,
# and the (x,y,z) position of the halo's center.

# to run this, set all the things below to your liking,
# and then 'python fastBuildMerge.py > mergeScript.py
# 'chmod u+x mergeScript.py'
# './mergeScript.py'

# the GraphViz file
outfile = "157-120.dot"
# the directory basename, no trailing slash needed
# if a data dump is /path/to/enzo/data/DD0243/data0243, put /path/to/enzo/data/DD below
dirbasename = "/path/to/enzo/data/DD"
# usually DD, maybe RD, or data
filebasename = "data"
# the maximum number of groups analyzed at one time step
hardmaxgroup = 2500
# the hop density threshold used for grouping
hopthreshold = 80.0
# dm only, True or False
dmonly = "True"

# The range of data dumps you wish to operate upon. It's a good idea to manually see how far back
# haloes exist before setting the end value.
# start > end
end = 120
start = 157

# A list of the haloes in the final data dump for which you want to find ancestors. Use the IDs provided
# by hop, so you'll probably have to run hop on the final data dump before deciding which haloes to
# put here, unless you know it's the largest halo, which is always ID=0. It can be a singleton.
indices = [0,1]

# change below to reflect where python is
# it is important to use python >= 2.5 as previous versions have a memory bug
# which affects this script, which is important if you're running large datasets.
print "#!/usr/bin/env /Library/Frameworks/Python.framework/Versions/Current/bin/python"

# ---------- you shouldn't have to change anything below. ---------
# -----------------------------------------------------------------
# -----------------------------------------------------------------

name = 'Merger'

print "import yt.lagos as lagos"
print "from yt.lagos import mylog"
print "from yt.mods import *"
print "import yt.lagos.hop.Merger as %s" % name

print """
def instantiateSnapshots(dirbasename,filebasename,start,end):
    redshift = {}
    snapshots = {}
    for i in range(start-end):
        j = start - i
        file = "%s%04d/%s%04d" % (dirbasename,j,filebasename,j)
        snapshot = lagos.EnzoStaticOutput(file)
        redshift[j] = snapshot["CosmologyCurrentRedshift"]
        # put the snapshot instance into the dict snapshots
        snapshots[j] = snapshot
    
    string = "Read in %d snapshots." % len(redshift)
    mylog.info(string)
    return(snapshots,redshift)
"""


print "positions = {}"
print "dirbasename = \"%s\"" % dirbasename
print "filebasename = \"%s\"" % filebasename
print "hardmaxgroup = %d" % hardmaxgroup
print "indices = " + str(indices)
print "start = %d" % start
print "end = %d" % end
print "dmonly = \"%s\"" % dmonly

# first let's read in the redshifts, and instantiate the snapshots
print "(snapshots,redshift) = instantiateSnapshots(dirbasename,filebasename,start,end-1)"

# open the dotfile
print "fopen = open('%s', 'w')" % outfile

# write the preambles for the dotfile
print "%s.writeTop(fopen)" % name
print "%s.writeNode(fopen)" % name

print "\n"

# loop over all the snapshots
for i in range(start-end):
    snapshot = start - i
    # snapshots should be in time-reverse order, so this loop will count *down*, so the next snapshot
    # to load has a lower index
    nextsnapshot = snapshot-1
    # if we're at the bottom, we're done!
    if (nextsnapshot < end):
        break
    
    # we need to run hop on the snapshot
    print "hop_results = HaloFinder.HOPHaloFinder(snapshots[%d], threshold=%f, dm_only=dmonly)" % \
        (nextsnapshot,hopthreshold)
    
    # get the group info for the next snapshot
    # get group positions
    print "positions = %s.convertPositions(hop_results,%d,positions,hardmaxgroup)" % (name,nextsnapshot)
    # get particle IDs
    print "g%04d = %s.convertGroups(hop_results,%d,hardmaxgroup)" % (nextsnapshot,name,nextsnapshot)
    # sort the particles in each group
    print "for g in g%04d: g.particles.sort()" % nextsnapshot
    # delete the sphere which contains all the particles, to clear up memory
    print "del sphere"
    # delete hop_results now that we're done with them, too
    print "del hop_results\n"
    
    # in the beginning we need to read in two groups before we build the links, so do everything above
    # again
    if (snapshot==start):
        print "hop_results = HaloFinder.HOPHaloFinder(snapshots[%d], threshold=%f, dm_only=dmonly)" % \
            (snapshot,hopthreshold)
        print "positions = %s.convertPositionsSelected(hop_results,%d,positions,indices)" % (name,snapshot)
        print "g%04d = %s.convertGroupsSelected(hop_results,%d,indices)" % (snapshot,name,snapshot)
        print "for g in g%04d: g.particles.sort()\n" % snapshot
        print "del sphere"
        print "del hop_results"
    
    
    # for just the last group (in time), a singleton, usually.
    if (snapshot==start):
        print "for g in g%04d: g.flag = 1\n" % start
    
    # build links which requires two neighboring groups at once.
    print "(links%04d%04d,g%04d,g%04d) = %s.buildLinks(g%04d,g%04d,positions)" % \
    (nextsnapshot,snapshot,nextsnapshot,snapshot,name,nextsnapshot,snapshot)
    
    # write out levels which keeps the boxes on one line rather than put how graphviz wants aesthetically
    
    print "%s.writeLevels(fopen, g%04d,redshift[%d])" % (name,nextsnapshot,nextsnapshot)
    if (snapshot==start):
        print "%s.writeLevels(fopen, g%04d,redshift[%d])" % (name,snapshot,snapshot)
    
    # print out label stuff
    
    print "%s.writeOpen(fopen)" % name
    
    # for the first (last in time) halo, write out the box labels using style 2
    if (snapshot==start):
        print "%s.writeLabels(fopen, g%04d, positions, 2)" % (name,snapshot)
    # middle time steps get type 1
    if (snapshot!=start):
        print "%s.writeLabels(fopen, g%04d, positions, 1)" % (name,snapshot)
    # the top, first in time get type 0
    if (nextsnapshot==end):
        print "%s.writeLabels(fopen, g%04d, positions, 0)" % (name,nextsnapshot)
    
    print "%s.writeClose(fopen)" % name
    
    print "%s.writeOpen(fopen)" % name
    print "%s.writeLinks(fopen,links%04d%04d, g%04d, g%04d)" % (name,nextsnapshot,snapshot,nextsnapshot,snapshot)
    print "%s.writeClose(fopen)" % name

    print "for g in g%04d: del g.particles\n\n" % snapshot

# close the dotfile
print "%s.writeClose(fopen)" % name
print "fopen.close()"
